import ellipsis as el
import datetime
import os
import time
import numpy as np
import pandas as pd

#python3 setup.py sdist bdist_wheel
#twine upload --repository pypi dist/*

folderId = '46e1e919-8b73-42a3-a575-25c6d45fd93b'

##account
token = el.account.logIn("demo_user", "")
admin_token = el.account.logIn('admin', "")
daan_token = el.account.logIn('daan', "")

el.account.listRootMaps('myDrive',token)
el.account.listRootFolders('sharedWithMe',token)

r_raster = el.path.searchRaster(token=token);

r_vector = el.path.searchVector()

r_raster = el.path.searchFolder(token=token);
    

##user

result = el.user.search('daan')
daanId = [x for x in result if x['username'] == 'daan'][0]['id']
el.user.get(daanId)


##path
rasterInfo = el.path.get('59caf510-bab7-44a8-b5ea-c522cfde4ad7', token)

info = el.path.get(folderId, token)

folderId = info['id']
maps = el.path.listMaps(folderId, token = token, listAll = True)
folders = el.path.listFolders(folderId, token = token, listAll = True)

mapId = [ m for m in maps['result'] if not m['trashed'] ][0]['id']


#crash


el.path.editMetadata(pathId = mapId, token = token, description = 'test')

addedFolderId = el.path.add( 'folder', 'test', token = token, parentId = folderId)['id']

el.path.trash(addedFolderId, token)
el.path.recover(addedFolderId, token)

addedRasterId = el.path.add( 'raster', 'test2', token = token, parentId = folderId)['id']

el.path.move([addedRasterId], addedFolderId, token)


el.path.trash(addedFolderId, admin_token)
el.path.delete(pathId = addedFolderId, token = admin_token, recursive = True)


el.path.editPublicAccess(pathId = folderId, token = token, accessLevel=0, hidden = False)
el.path.editPublicAccess(pathId = folderId, token = token, accessLevel=100, hidden = True)

el.path.favorite(folderId, token=token)
el.path.unfavorite(folderId, token=token)

###invites
inviteId =  el.path.invite.send(pathId = folderId, token=token, userId = daanId, access = {'accessLevel': 200, 'processingUnits':10000})['id']

el.path.invite.getPathInvites(folderId, token)


el.path.invite.getYourInvites(daan_token)

el.path.invite.revoke(pathId = folderId, inviteId = inviteId, token = token)


inviteId =  el.path.invite.send(pathId = folderId, token=token, userId = daanId, access = {'accessLevel': 200, 'processingUnits':10000})['id']

el.path.invite.decline(folderId, inviteId, daan_token)


inviteId =  el.path.invite.send(pathId = folderId, token=token, userId = daanId, access = {'accessLevel': 200, 'processingUnits':10000})['id']
el.path.invite.accept(folderId, inviteId, daan_token)

##members
members = el.path.member.get(folderId, token, memberType = ['direct'])


daanMemberId = [m['user']['id'] for m in members if m['user']['username'] == 'daan'][0]

el.path.member.edit(pathId = folderId, userId = daanMemberId, access = {'accessLevel' : 200} ,token = token)

el.path.member.delete(folderId, daanMemberId, token)


##raster and uploads
mapId = el.path.add('raster', 'test', token, parentId = folderId)['id']
el.path.raster.editMap(mapId, token, interpolation = 'nearest')


timestampId = el.path.raster.timestamp.add(mapId, token)['id']


dateFrom = datetime.datetime.now()
dateTo = datetime.datetime.now()
el.path.raster.timestamp.edit(mapId, timestampId, token, description = 'hoi', dateFrom = dateFrom, dateTo = dateTo)

filePath = 'test/0.tif'
uploadId = el.path.raster.timestamp.upload.upload(mapId, timestampId, filePath, token)['id']

el.path.raster.timestamp.upload.delete(mapId, timestampId, uploadId, token)
uploadId = el.path.raster.timestamp.upload.upload(mapId, timestampId, filePath, token)['id']

uploads = el.path.raster.timestamp.upload.get(mapId, timestampId, token)

el.path.raster.timestamp.activate(mapId, timestampId, token)

info = el.path.get(mapId, token)
timestamp = info['raster']['timestamps'][0]
while timestamp['status'] != 'finished':
    time.sleep(1)
    info = el.path.get(mapId, token)
    timestamp = info['raster']['timestamps'][0]


el.path.raster.editBand(mapId, 1, 'hoi', token)


el.path.raster.timestamp.trash(mapId, timestampId, token)
el.path.raster.timestamp.recover(mapId, timestampId, token)
el.path.raster.timestamp.trash(mapId, timestampId, token)
el.path.raster.timestamp.delete(mapId, timestampId, admin_token)

##raster information retrieval
mapId = '59caf510-bab7-44a8-b5ea-c522cfde4ad7'
timestampId = 'f25e120e-ca8f-451f-a5f4-33791db0f2c5'


xMin  = 5.60286
yMin=  52.3031    
xMax  = 5.60315
yMax  = 52.30339

extent = {'xMin':xMin,'yMin':yMin,'xMax':xMax,'yMax':yMax } 

result = el.path.raster.timestamp.getRaster(pathId = mapId, timestampId = timestampId, extent = extent, token = token)

raster = result['raster']


el.util.plotRaster(raster[0:3,:,:])

r = el.path.raster.timestamp.getDownsampledRaster(pathId = mapId, timestampId=timestampId, extent = extent, width = 256, height = 256, token = token)
raster = r['raster']
el.util.plotRaster(raster[0:3,:,:])


bounds = el.path.raster.timestamp.getBounds(mapId, timestampId, token)

data = el.path.raster.timestamp.getAggregatedData(mapId, [timestampId], bounds, token=token)

###raster downloads
mapId = '1eea3d2f-27b3-4874-b716-87852c3407c1'
timestampId = "ba5b418a-a39e-4d84-9411-e23c096085a3"
uploads = el.path.raster.timestamp.upload.get(mapId, timestampId, token)
uploadId = uploads['result'][0]['id']

downloadId = el.path.raster.timestamp.order.order(mapId, timestampId, token, uploadId = uploadId)['id']

file_out = '/home/daniel/Downloads/out.tif'
el.path.raster.timestamp.order.download(downloadId, file_out, token)
os.remove(file_out)

pendinDownloads = el.path.raster.timestamp.order.get(token)

parameters = {"angle":45,"bandNumber":1,"alpha":1}
method = "hillShade"
layerId = el.path.raster.layer.add(mapId, 'hoi', method, parameters, token, 'hoi')['id']

el.path.raster.layer.edit(mapId, layerId, method= method, parameters = parameters,token = token, description =  'doei')
el.path.raster.layer.delete(mapId, layerId, token)



###vector layers
mapId = el.path.add('vector', 'test', token)['id']


layerId = el.path.vector.layer.add(mapId, 'test', token)['id']

el.path.vector.layer.edit(mapId, layerId, token, name = 'test2')

el.path.vector.layer.archive(mapId, layerId, token)
el.path.vector.layer.recover(mapId, layerId, token)
el.path.vector.layer.archive(mapId, layerId, token)
el.path.vector.layer.delete(mapId, layerId, admin_token)
layerId = el.path.vector.layer.add(mapId, 'test', token)['id']


###vector uploads
filePath = 'test/test.zip'
el.path.vector.layer.upload.upload(mapId, layerId, filePath, token, fileFormat = 'zip')


upload = el.path.vector.layer.upload.get(mapId, layerId, token)['result'][0]

while upload['status'] != 'completed':
    time.sleep(1)
    upload = el.path.vector.layer.upload.get(mapId, layerId, token)['result'][0]


#layer methods
bounds = el.path.vector.layer.getBounds(mapId, layerId, token)

xMin = bounds.bounds[0]
yMin = bounds.bounds[1]
xMax = bounds.bounds[2]
yMax = bounds.bounds[3]

bounds = {'xMin':xMin, 'xMax':xMax, 'yMin':yMin, 'yMax':yMax}
sh = el.path.vector.layer.getFeaturesByExtent(mapId, layerId, bounds)
sh['result'].plot()

sh = el.path.vector.layer.listFeatures(mapId, layerId, token)

featureIds = sh['result']['id'].values

r =  el.path.vector.layer.getFeaturesByIds(mapId, layerId, featureIds, token)


r = el.path.vector.layer.getChanges(mapId, layerId, token, listAll = True)

el.path.vector.layer.editFilter(mapId, layerId, [{'property':'gml_id'}], token)

r = el.path.get(mapId, token)
blocked = r['vector']['layers'][0]['availability']['blocked']
while blocked:
    time.sleep(1)
    r = el.path.get(mapId, token)
    blocked = r['vector']['layers'][0]['availability']['blocked']


###feature module
features = sh['result'][1:2]
featureIds = features['id'].values

el.path.vector.layer.feature.add(mapId, layerId, features, token)


el.path.vector.layer.feature.edit(mapId, layerId, featureIds = features['id'].values, token = token, features = features)

el.path.vector.layer.feature.delete(mapId, layerId, featureIds, token)
el.path.vector.layer.feature.recover(mapId, layerId, featureIds, token)
featureId = featureIds[0]
el.path.vector.layer.feature.versions(mapId, layerId, featureId, token)


###message module
el.path.vector.layer.feature.message.add(mapId, layerId, featureId, token, text= 'hoi')
image = np.zeros((256,256))
el.path.vector.layer.feature.message.add(mapId, layerId, featureId, token, text= 'hoi', image = image)

messages = el.path.vector.layer.feature.message.get(mapId, layerId, featureIds=[featureId], token = token)

messageId = [m for m in messages['result'] if m['thumbnail'] != None][0]['id']


el.path.vector.layer.feature.message.getImage(mapId, layerId, messageId, token)

el.path.vector.layer.feature.message.delete(mapId, layerId, messageId, token)
el.path.vector.layer.feature.message.recover(mapId, layerId, messageId, token)


###series module
date = datetime.datetime.now()
seriesData = pd.DataFrame({'x': [1,2,3,4]})
seriesData['date'] = date
el.path.vector.layer.feature.series.add(pathId = mapId, layerId = layerId, featureId = featureId, seriesData = seriesData, token = token)


el.path.vector.layer.feature.series.info(mapId, layerId, featureId,token)

r = el.path.vector.layer.feature.series.get(mapId, layerId, featureId, token = token)


seriesId = r['result']['id'].values[0]

el.path.vector.layer.feature.series.delete(mapId, layerId, featureId, [seriesId], token)

el.path.vector.layer.feature.series.recover(mapId, layerId, featureId, [seriesId], token)

el.path.vector.layer.feature.series.changelog(mapId, layerId, featureId)



#style module

parameters = {"alpha":0.5,"width":2,"radius":{"method":"constant","parameters":{"value":7}},"property":"gml_id"}
styleId = el.path.vector.layer.style.add(mapId, layerId, 'test', 'random', parameters, token)['id']

el.path.vector.layer.style.edit(mapId, layerId, styleId, token, name = 'sfd')
el.path.vector.layer.style.delete(mapId, layerId, styleId, token)


## properties module
featurePropertyId = el.path.vector.layer.featureProperty.add(mapId, layerId, 'new', 'string', token)['id']
el.path.vector.layer.featureProperty.delete(mapId, layerId, featurePropertyId, token)

el.path.vector.layer.featureProperty.recover(mapId, layerId, featurePropertyId, token)
el.path.vector.layer.featureProperty.edit(mapId, layerId, featurePropertyId, token, required = True)

### order module

orderId = el.path.vector.layer.order.order(mapId, layerId, token, bounds)['id']

order = el.path.vector.layer.order.get(token)[0]
while order['status'] != 'completed':
    time.sleep(1)
    order = el.path.vector.layer.order.get(token)[0]

file_out = '/home/daniel/Downloads/out.zip'
el.path.vector.layer.order.download(orderId, file_out, token)
os.remove(file_out)    
el.path.trash(mapId, token)









from ellipsis.util.root import plotRaster
from ellipsis.util.root import plotFeatures
from ellipsis.util.root import chunks
from ellipsis.util.root import cover
from ellipsis.util.root import loadingBar

from ellipsis.util.root import reprojectRaster
from ellipsis.util.root import saveRaster
from ellipsis.util.root import transformPoint

__version__ = '1.3.0'






from ellipsis import apiManager
from ellipsis import sanitize
from ellipsis.util.root import recurse
import os

def upload(pathId, timestampId, filePath, token, epsg = None, noDataValue = None, fileFormat = 'tif'):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    filePath = sanitize.validString('filePath', filePath, True)
    noDataValue = sanitize.validFloat('noDataValue', noDataValue, False)    
    epsg = sanitize.validInt('epsg', epsg, False)    
    fileFormat = sanitize.validString('fileFormat', fileFormat, False)    
        
    seperator = os.path.sep    
    fileName = filePath.split(seperator)[len(filePath.split(seperator))-1 ]
    
    body = {'fileName':fileName, 'epsg':epsg, 'noDataValue':noDataValue, 'format':fileFormat}
    r = apiManager.upload('/path/' + pathId + '/raster/timestamp/' + timestampId + '/upload' , filePath, body, token)
    return r

def get(pathId, timestampId, token, pageStart= None, listAll = True):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 
    pageStart = sanitize.validUuid('pageStart', pageStart, False) 

    def f(body):
        r = apiManager.get('/path/' + pathId + '/raster/timestamp/' + timestampId + '/upload', None, token)    
        return r

    r = recurse(f, {'pageStart':pageStart}, listAll)
    return r
    

def delete(pathId, timestampId, uploadId, token):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True) 
    uploadId = sanitize.validUuid('uploadId', uploadId, True) 
    timestampId = sanitize.validUuid('timestampId', timestampId, True) 

    r = apiManager.delete('/path/' + pathId + '/raster/timestamp/' + timestampId + '/upload/' + uploadId, None, token)    
    return r




from ellipsis import apiManager
from ellipsis import sanitize
from ellipsis.util import chunks
from ellipsis.util import loadingBar
from ellipsis.util.root import stringToDate
from ellipsis.util.root import recurse

import numpy as np
import json
import geopandas as gpd


def add(pathId, layerId, features, token, zoomlevels = None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    token = sanitize.validString('token', token, True)
    features = sanitize.validGeopandas('features', features, True)
    zoomlevels = sanitize.validIntArray('zoomlevels', zoomlevels, False)
        
    
    #check if first time
    firstTime = apiManager.get('/path/' + pathId, None, token)
    firstTime = [x for x in firstTime['vector']['layers'] if x['id'] == layerId]
    if len(firstTime)==0:
        raise ValueError('layer does not exist')
    firstTime = len(firstTime[0]['properties']) ==0
    
    if firstTime:
        print('no properties known for this layer adding them automatically')
        columns = features.columns
        columns = [c for c in columns if c != 'geometry']
        for c in columns:
            if 'int' in str(features.dtypes[c]) or 'Int' in str(features.dtypes[c]):
                propertyType = 'integer'
                features[c] = [ int(d) if not np.isnan(d) and d != None else  np.nan for d in features[c].values ]
            elif 'float' in str(features.dtypes[c]) or 'Float' in str(features.dtypes[c]):
                propertyType = 'float'
                features[c] = [ float(d) if not np.isnan(d) and d != None else  np.nan for d in features[c].values ]
            elif 'bool' in str(features.dtypes[c]):
                propertyType = 'boolean'
                features[c] = [ bool(d) if not np.isnan(d) and d != None else  np.nan for d in features[c].values ]
            elif 'datetime' in str(features.dtypes[c]):
                propertyType = 'datetime'
                features[c] = [ d.strftime('%Y-%m-%d %H:%M:%S.%f')[:-3] for d in features[c].values ]
            else:
                propertyType = 'string'
                features[c] = [ str(d) if d != None else  np.nan for d in features[c].values ]


            ###date
            body = {'name': c , 'type': propertyType , 'required': False, 'private': False}
            apiManager.post('/path/' + pathId + '/vector/layer/' + layerId + '/property', body, token)                
    indices = chunks(np.arange(features.shape[0]))


    addedIds = []
    for i in np.arange(len(indices)):
        indices_sub = indices[i]
        features_sub = features.iloc[indices_sub]
        features_sub =features_sub.to_json(na='drop')
        features_sub = json.loads(features_sub)
        
        body = {"features":features_sub['features'], 'zoomlevels':zoomlevels}

        r = apiManager.post('/path/' + pathId + '/vector/layer/' + layerId + '/feature', body, token)

        addedIds = addedIds + r

        loadingBar(i*3000 + len(indices_sub),features.shape[0])
        i = i+1
        
    return(addedIds)


    
def edit(pathId, layerId, featureIds, token, zoomlevels = None, features = None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    token = sanitize.validString('token', token, True)
    features = sanitize.validGeopandas('features', features, False)
    zoomlevels = sanitize.validIntArray('zoomlevels', zoomlevels, False)
    featureIds = sanitize.validUuidArray('featureIds', featureIds, True)


    if type(features) != type(None) and features.shape[0] != len(featureIds):
        raise ValueError('featureIds must be of same length as the features geopandas dataframe')
        

    indices = chunks(np.arange(len(featureIds)),1000)
    i=0
    for i in np.arange(len(indices)):
        indices_sub = indices[i]
        featureIds_sub = list( np.array(featureIds)[indices_sub])

        if type(features) != type(None):
            features_sub = features.iloc[indices_sub]        
            features_sub =features_sub.to_json(na='drop')
            features_sub = json.loads(features_sub)

        if str(type(zoomlevels)) != str(type(None)) and str(type(features)) != str(type(None)):
            changes = [{'featureId':x[0] , 'newProperties':x[1]['properties'], 'newGeometry':x[1]['geometry'], 'newZoomlevels':zoomlevels} for x in zip(featureIds_sub, features_sub['features'])]
        elif str(type(zoomlevels)) != str(type(None)) and str(type(features)) == str(type(None)):
            changes = [{'featureId':geometryId, 'newZoomlevels':zoomlevels} for geometryId in featureIds]
        else:
            changes = [{'featureId':x[0] , 'newProperties':x[1]['properties'], 'newGeometry':x[1]['geometry']} for x in zip(featureIds_sub, features_sub['features'])]
            
        body = {'changes':changes}
        r = apiManager.patch('/path/' + pathId + '/vector/layer/' + layerId + '/feature', body, token)

        if len(indices) > 1:
            loadingBar(i*1000 + len(indices_sub),len(featureIds))

    return r


def delete(pathId, layerId, featureIds, token):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    token = sanitize.validString('token', token, True)
    featureIds = sanitize.validUuidArray('featureIds', featureIds, True)

    indices = chunks(np.arange(len(featureIds)),1000)
    i=0
    for i in np.arange(len(indices)):
        indices_sub = indices[i]
        featureIds_sub = list(np.array(featureIds)[indices_sub])

        body = {'featureIds': featureIds_sub, 'deleted': True}
        r = apiManager.put('/path/' + pathId + '/vector/layer/' + layerId + '/feature/deleted', body, token)

        if len(indices) > 1:
            loadingBar(i*1000 + len(indices_sub),len(featureIds))

    return r


def recover(pathId, layerId, featureIds, token):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    token = sanitize.validString('token', token, True)
    featureIds = sanitize.validUuidArray('featureIds', featureIds, True)

    indices = chunks(np.arange(len(featureIds)),1000)
    i=0
    for i in np.arange(len(indices)):
        indices_sub = indices[i]
        featureIds_sub = list(np.array(featureIds)[indices_sub])

        body = {'featureIds': featureIds_sub, 'deleted': False}
        r = apiManager.put('/path/' + pathId + '/vector/layer/' + layerId + '/feature/deleted', body, token)

        if len(indices) > 1:
            loadingBar(i*1000 + len(indices_sub),len(featureIds))

    return r



def versions(pathId, layerId, featureId, token = None, pageStart = None, listAll = True):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    token = sanitize.validString('token', token, False)
    featureId = sanitize.validUuid('featureId', featureId, True)
    pageStart = sanitize.validUuid('pageStart', pageStart, False)
    listAll = sanitize.validBool('listAll', listAll, True)
    
    body = {'returnType':'all'}
    def f(body):
        return apiManager.get('/path/' + pathId + '/vector/layer/' + layerId + '/feature/' + featureId + '/version', body, token)

    r = recurse(f, body, listAll)

    features = [ x['feature'] for x in r['result']]
    dates = [stringToDate(x['date']) for x in r['result'] ]
    usernames = [x['user']['username'] for x in r['result'] ]
    userIds = [x['user']['id'] for x in r['result'] ]

    sh = gpd.GeoDataFrame.from_features(features)     
    sh['username'] = usernames
    sh['userId'] = userIds
    sh['dates'] = dates
    
    sh.crs = {'init': 'epsg:4326'}
    r['result'] = sh
    return(r)





from ellipsis import sanitize
from ellipsis import apiManager
from ellipsis.util import loadingBar
from ellipsis.util import chunks
from ellipsis.util.root import transformPoint

import json
import numpy as np
from io import BytesIO
import rasterio
import math
import tifffile
import threading
from PIL import Image
import geopandas as gpd
import datetime

def getDownsampledRaster(pathId, timestampId, extent, width, height, layer = None, token = None):
    bounds = extent
    token = sanitize.validString('token', token, False)
    pathId = sanitize.validUuid('pathId', pathId, True)
    timestampId = sanitize.validUuid('timestampId', timestampId, True)
    bounds = sanitize.validBounds('bounds', bounds, True)
    layer = sanitize.validObject('layer', layer, False)

    body = {'pathId':pathId, 'timestampId':timestampId, 'bounds':bounds, 'width':width, 'height':height}
    r = apiManager.get('/path/' + pathId + '/raster/timestamp/' + timestampId + '/rasterByBounds', body, token, crash = False)
    if r.status_code != 200:
        raise ValueError(r.message)


    if type(layer) == type(None):
        r = tifffile.imread(BytesIO(r.content))
    else:
        r = np.array(Image.open(BytesIO(r.content)))
        r = np.transpose(r, [2,0,1])

    xMin = bounds['xMin']
    yMin = bounds['yMin']
    xMax = bounds['xMax']
    yMax = bounds['yMax']

    xMinWeb, yMinWeb = transformPoint( (xMin, yMin), 'EPSG:4326', 'EPSG:3857')
    xMaxWeb, yMaxWeb = transformPoint( (xMax, yMax), 'EPSG:4326', 'EPSG:3857')


    trans = rasterio.transform.from_bounds(xMinWeb, yMinWeb, xMaxWeb, yMaxWeb, r.shape[2], r.shape[1])

    return {'raster': r, 'transform':trans, 'extent': {'xMin' : xMinWeb, 'yMin': yMinWeb, 'xMax': xMaxWeb, 'yMax': yMaxWeb}, 'crs':"EPSG:3857"}




def getRaster(pathId, timestampId, extent, layer = None, threads = 1, token = None):
    bounds = extent
    threads = sanitize.validInt('threads', threads, True)
    token = sanitize.validString('token', token, False)
    pathId = sanitize.validUuid('pathId', pathId, True)
    timestampId = sanitize.validUuid('timestampId', timestampId, True)
    bounds = sanitize.validBounds('bounds', bounds, True)
    layer = sanitize.validObject('layer', layer, False)

        
    xMin = bounds['xMin']
    yMin = bounds['yMin']
    xMax = bounds['xMax']
    yMax = bounds['yMax']
    xMinWeb,yMinWeb =  transformPoint((xMin, yMin), 'EPSG:4326', 'EPSG:3857')
    xMaxWeb,yMaxWeb = transformPoint((xMax, yMax), 'EPSG:4326', 'EPSG:3857')

    info = apiManager.get('/path/' + pathId, None, token)
    bands =  info['raster']['bands']
    if type(layer) == type(None):
        num_bands = len(bands)
        dtype = info['raster']['format']
    else:
        num_bands = 4
        dtype = 'uint8'

    timestamps =  info['raster']['timestamps']
    all_timestamps = [item['id'] for item in timestamps]
    if not timestampId in all_timestamps:
        raise ValueError('given timestamp does not exist')
    
    zoom = next(item for item in timestamps if item["id"] == timestampId)['zoom']

    body = {'layer':layer}

    LEN = 2.003751e+07

    x_start = 2**zoom * (xMinWeb + LEN) / (2* LEN)
    x_end = 2**zoom * (xMaxWeb + LEN) / (2* LEN)
    y_end = 2**zoom * (LEN - yMinWeb) / (2* LEN)
    y_start = 2**zoom * (LEN - yMaxWeb) / (2* LEN)

    x1_osm = math.floor(x_start)
    x2_osm = math.floor(x_end)
    y1_osm = math.floor(y_start)
    y2_osm = math.floor(y_end)
    
    x_tiles = np.arange(x1_osm, x2_osm+1)
    y_tiles = np.arange(y1_osm, y2_osm +1)
    
    r_total = np.zeros((num_bands, 256*(y2_osm - y1_osm + 1) ,256*(x2_osm - x1_osm + 1)), dtype = dtype)
    
    tiles = []            
    for tileY in y_tiles:
        for tileX in x_tiles:
            tiles = tiles + [(tileX, tileY)]
    def subTiles(tiles):
            N = 0
            for tile in tiles:
                tileX = tile[0]
                tileY = tile[1]
                x_index = tileX - x1_osm
                y_index = tileY - y1_osm
                
                r = apiManager.get('/path/' + pathId + '/raster/timestamp/' + timestampId + '/tile/' + str(zoom) + '/' + str(tileX) + '/' + str(tileY), body, token, False)

                if r.status_code == 403:
                        raise ValueError('insufficient access')
                if r.status_code != 200:
                        r = np.zeros((num_bands,256,256))
                else:
                    print('tile found')
                    if type(layer) == type(None):
                        r = tifffile.imread(BytesIO(r.content))
                    else:
                        r = np.array(Image.open(BytesIO(r.content)))
                        r = np.transpose(r, [2,0,1])

                r = r.astype(dtype)
                r_total[:,y_index*256:(y_index+1)*256,x_index*256:(x_index+1)*256] = r
                loadingBar(N, len(tiles))
                N = N + 1
            

    size = math.floor(len(tiles)/threads) + 1
    tiles_chunks = chunks(tiles, size)
    prs = []
    for tiles in tiles_chunks:
        pr = threading.Thread(target = subTiles, args =(tiles,), daemon=True)
        pr.start()
        prs = prs + [pr]
    for pr in prs:
        pr.join()
        
    min_x_index = int(math.floor((x_start - x1_osm)*256))
    max_x_index = max(int(math.floor((x_end- x1_osm)*256 + 1 )), min_x_index + 1 )
    min_y_index = int(math.floor((y_start - y1_osm)*256))
    max_y_index = max(int(math.floor((y_end- y1_osm)*256 +1)), min_y_index + 1)

    r_total = r_total[:,min_y_index:max_y_index,min_x_index:max_x_index]
 
    trans = rasterio.transform.from_bounds(xMinWeb, yMinWeb, xMaxWeb, yMaxWeb, r_total.shape[2], r_total.shape[1])

    return {'raster': r_total, 'transform':trans, 'extent': {'xMin' : xMinWeb, 'yMin': yMinWeb, 'xMax': xMaxWeb, 'yMax': yMaxWeb}, 'crs':"EPSG:3857"}


def getAggregatedData(pathId, timestampIds, geometry, approximate=True, token = None):
    token = sanitize.validString('token', token, False)
    pathId = sanitize.validUuid('pathId', pathId, True)    
    timestampIds = sanitize.validUuidArray('timestampIds', timestampIds, True)    
    approximate = sanitize.validBool(approximate, approximate, True)    
    geometry = sanitize.validShapely('geometry', geometry, True)
    try:
        sh = gpd.GeoDataFrame({'geometry':[geometry]})
        geometry =sh.to_json(na='drop')
        geometry = json.loads(geometry)
        geometry = geometry['features'][0]['geometry']
    except:
        raise ValueError('geometry must be a shapely geometry')


    body = {'timestampIds':timestampIds, 'geometry':geometry, 'approximate':approximate}
    r = apiManager.get('/path/' + pathId + '/raster/timestamp/analyse', body, token)
    return r


def add(pathId, token, dateFrom = None, dateTo = None, description= None,  appendToTimestampId = None):
    
    if type(dateFrom) == type(None):
        dateFrom = datetime.datetime.now()
    if type(dateTo) == type(None):
        dateTo = datetime.datetime.now()


    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    dateFrom = sanitize.validDate('dateFrom', dateFrom, True)    
    dateTo = sanitize.validDate('dateTo', dateTo, True)    
    description = sanitize.validString('description', description, False)    
    appendToTimestampId = sanitize.validUuid('appendToTimestampId', appendToTimestampId, False)    

    body = {'dateFrom' : dateFrom, 'dateTo':dateTo, 'description': description, 'appendToTimestampId': appendToTimestampId}    
    r = apiManager.post('/path/' + pathId + '/raster/timestamp'  , body, token)
    return r
    
def edit(pathId, timestampId, token, dateFrom = None, dateTo = None, description= None):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    timestampId = sanitize.validUuid('timestampId', timestampId, True)  
    dateFrom = sanitize.validDate('dateFrom', dateFrom, False)    
    dateTo = sanitize.validDate('dateTo', dateTo, False)    
    description = sanitize.validString('description', description, False)    

    body = {'dateFrom' : dateFrom, 'dateTo':dateTo, 'description': description}    
    r = apiManager.patch('/path/' + pathId + '/raster/timestamp/' + timestampId  , body, token)
    return r

def getBounds(pathId, timestampId, token = None):
    token = sanitize.validString('token', token, False)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    timestampId = sanitize.validUuid('timestampId', timestampId, True)  
    r = apiManager.get('/path/' + pathId + '/raster/timestamp/' + timestampId + '/bounds'  , None, token)
    r['id'] = 0
    r['properties'] = {}
    r  = gpd.GeoDataFrame.from_features([r])
    r = r.unary_union
    return r

def activate(pathId, timestampId, token):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    timestampId = sanitize.validUuid('timestampId', timestampId, True)  
    r = apiManager.post('/path/' + pathId + '/raster/timestamp/' + timestampId + '/activate'  , None, token)
    return r

def delete(pathId, timestampId, token):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    timestampId = sanitize.validUuid('timestampId', timestampId, True)  

    r = apiManager.delete('/path/' + pathId + '/raster/timestamp/' + timestampId  , None, token)
    return r

def trash(pathId, timestampId, token):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    timestampId = sanitize.validUuid('timestampId', timestampId, True)  
    body = {'trashed' : True}
    r = apiManager.put('/path/' + pathId + '/raster/timestamp/' + timestampId + '/trashed'  , body, token)
    return r

    
def recover(pathId, timestampId, token):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)  
    timestampId = sanitize.validUuid('timestampId', timestampId, True)  
    body = {'trashed' : False}
    r = apiManager.put('/path/' + pathId + '/raster/timestamp/' + timestampId + '/trashed'  , body, token)
    return r



from ellipsis import apiManager
from ellipsis import sanitize
import os
from ellipsis.util.root import recurse

def upload(pathId, layerId, filePath, token, epsg = None, fileFormat = 'geojson', dateColumns = None, datePatterns = None):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    filePath = sanitize.validString('filePath', filePath, True)
    epsg = sanitize.validInt('epsg', epsg, False)    
    fileFormat = sanitize.validString('fileFormat', fileFormat, False)    
    dateColumns = sanitize.validStringArray('dateColumns', dateColumns, False)    
    datePatterns = sanitize.validStringArray('datePatterns', datePatterns, False)    
        
    seperator = os.path.sep    
    fileName = filePath.split(seperator)[len(filePath.split(seperator))-1 ]
    
    body = {'fileName':fileName, 'epsg':epsg, 'format':fileFormat, 'dateColumns': dateColumns, 'datePatterns':datePatterns}
    apiManager.upload('/path/' + pathId + '/vector/layer/' + layerId + '/upload' , filePath, body, token)


def get(pathId, layerId, token, pageStart = None, listAll = True):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    pageStart = sanitize.validUuid('pageStart', pageStart, False)
    listAll = sanitize.validBool('listAll', listAll, True)

    body = {'pageStart': pageStart}
    def f(body):    
        r = apiManager.get('/path/' + pathId + '/vector/layer/' + layerId + '/upload', body, token)    
        return r

    r = recurse(f, body, listAll)
    return r


from ellipsis import apiManager
from ellipsis import sanitize
from ellipsis.util.root import recurse
import geopandas as gpd

from ellipsis.util import chunks
from ellipsis.util import loadingBar
from ellipsis.util.root import stringToDate

def add(pathId, name, token, properties = None, description = None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    name = sanitize.validString('name', name, True)
    token = sanitize.validString('token', token, True)
    properties = sanitize.validObject('properties', properties, False)
    description = sanitize.validString('description', description, False)

    body = {'name':name, 'properties':properties, 'description':description}
    r = apiManager.post('/path/' + pathId + '/vector/layer', body, token)
    return r

def edit(pathId, layerId, token, description=None, name=None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    name = sanitize.validString('name', name, True)
    token = sanitize.validString('token', token, True)
    description = sanitize.validString('description', description, False)

    body = {'name':name,'description':description}
    r = apiManager.patch('/path/' + pathId + '/vector/layer/' + layerId, body, token)
    return r
    

def archive(pathId, layerId, token):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    token = sanitize.validString('token', token, True)
    body = {'trashed':True}
    r = apiManager.put('/path/' + pathId + '/vector/layer/' + layerId + '/trashed', body, token)
    return r


def recover(pathId, layerId, token):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    token = sanitize.validString('token', token, True)
    body = {'trashed':False}
    r = apiManager.put('/path/' + pathId + '/vector/layer/' + layerId + '/trashed', body, token)
    return r

def delete(pathId, layerId, token):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    token = sanitize.validString('token', token, True)
    r = apiManager.delete('/path/' + pathId + '/vector/layer/' + layerId , None, token)
    return r
    
    
def getBounds(pathId, layerId, token = None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    token = sanitize.validString('token', token, False)
    r = apiManager.get('/path/' + pathId + '/vector/layer/' + layerId + '/bounds' , None, token)

    r['id'] = 0
    r['properties'] = {}
    r  = gpd.GeoDataFrame.from_features([r])
    r = r.unary_union

    return r


def getChanges(pathId, layerId, token = None, pageStart = None, listAll = False, actions = None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    token = sanitize.validString('token', token, False)
    listAll = sanitize.validBool('listAll', listAll, True)
    pageStart = sanitize.validObject('pageStart', pageStart, False) 
    actions = sanitize.validObject('actions', actions, False)
    body = {'pageStart':pageStart}    
    def f(body):
        r = apiManager.get('/path/' + pathId + '/vector/layer/' + layerId + '/changelog' , body, token)
        return r
    
    r = recurse(f, body, listAll)
    r['result'] = [ {**x, 'date':stringToDate(x['date'])} for x in r['result'] ]
    return r

def editFilter(pathId, layerId, propertyFilter, token):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    token = sanitize.validString('token', token, True)
    propertyFilter = sanitize.validObject('propertyFilter', propertyFilter, True)
    
    body = {'filter': propertyFilter}
    r = apiManager.post('/path/' + pathId + '/vector/layer/' + layerId + '/filter' , body, token)
    return r


def getFeaturesByIds(pathId, layerId, featureIds, token = None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    token = sanitize.validString('token', token, False)
    featureIds = sanitize.validUuidArray('featureIds', featureIds, True)
    
    id_chunks = chunks(featureIds, 10)

    r = {'size': 0 , 'result': [], 'nextPageStart' : None}
    ids = id_chunks[0]
    i=0
    for ids in id_chunks:
        body = {'geometryIds': ids}
        r_new = apiManager.get('/path/' + pathId + '/vector/layer/' + layerId + '/featuresByIds' , body, token)
        
        r['result'] = r['result'] + r_new['result']['features']
        r['size'] = r['size'] + r_new['size']
        if len(id_chunks) >0:
            loadingBar(i*10 + len(ids),len(featureIds))
        i=i+1

        
    sh = gpd.GeoDataFrame.from_features(r['result'])    
    r['result'] = sh
    return r
    

def getFeaturesByExtent(pathId, layerId, extent, propertyFilter = None, token = None, listAll = True, pageStart = None):
    bounds = extent
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    token = sanitize.validString('token', token, False)
    bounds = sanitize.validBounds('bounds', bounds, True)
    propertyFilter = sanitize.validObject('propertyFilter', propertyFilter, False)
    listAll = sanitize.validBool('listAll', listAll, True)
    pageStart = sanitize.validUuid('pageStart', pageStart, False) 
    
    body = {'pageStart': pageStart, 'propertyFilter':propertyFilter, 'bounds':bounds}

    def f(body):
        return apiManager.get('/path/' + pathId + '/vector/layer/' + layerId + '/featuresByBounds' , body, token)
        
    r = recurse(f, body, listAll, 'features')

    sh = gpd.GeoDataFrame.from_features(r['result']['features'])
    r['result'] = sh
    return r


def listFeatures(pathId, layerId, token = None, listAll = True, pageStart = None):
    pathId = sanitize.validUuid('pathId', pathId, True) 
    layerId = sanitize.validUuid('layerId', layerId, True) 
    token = sanitize.validString('token', token, False)
    listAll = sanitize.validBool('listAll', listAll, True)
    pageStart = sanitize.validUuid('pageStart', pageStart, False) 

    body = {'pageStart': pageStart}

    def f(body):
        return apiManager.get('/path/' + pathId + '/vector/layer/' + layerId + '/listFeatures' , body, token)

    r = recurse(f, body, listAll, 'features')

    
    sh = gpd.GeoDataFrame.from_features(r['result']['features'])    
    r['result'] = sh

    return r







from ellipsis import apiManager
from ellipsis import sanitize

def add(pathId, name, method, parameters, token, description = None):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)    
    method = sanitize.validString('method', method, True)    
    name = sanitize.validString('name', name, True)    
    description = sanitize.validString('description', description, False)    
    parameters = sanitize.validObject('parameters', parameters, True)

    body = {'name':name, 'method': method, 'parameters':parameters, 'description': description}
    r = apiManager.post('/path/' + pathId + '/raster/layer', body, token)
    return r

    
def delete(pathId, layerId, token):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)    
    layerId = sanitize.validUuid('layerId', layerId, True)    

    r = apiManager.delete('/path/' + pathId + '/raster/layer/' + layerId, None, token)
    return r


def edit(pathId, layerId, method, parameters, token, description = None):
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)    
    method = sanitize.validString('method', method, True)    
    description = sanitize.validString('description', description, False)    
    parameters = sanitize.validObject('parameters', parameters, True)

    body = {'method': method, 'parameters':parameters, 'description': description}
    r = apiManager.patch('/path/' + pathId + '/raster/layer/' + layerId, body, token)
    return r





from ellipsis import sanitize
from ellipsis import apiManager

def get(token):
    r = apiManager.get('/path/vector/layer/order', None, token)
    return r

def order(pathId, layerId, token, bounds = None, uploadId = None):
    
    token = sanitize.validString('token', token, True)
    pathId = sanitize.validUuid('pathId', pathId, True)
    layerId = sanitize.validUuid('layerId', layerId, True)
    bounds = sanitize.validBounds('bounds', bounds, False)
    uploadId = sanitize.validUuid('uploadId', uploadId, False)

    body = {'uploadId': uploadId, 'bounds':bounds, 'format' :'geojson'}
    
    r = apiManager.post('/path/' + pathId + '/vector/layer/' + layerId + '/order', body, token)    

    return r  

def download(orderId, filePath, token):
    
    token = sanitize.validString('token', token, True)
    orderId = sanitize.validUuid('orderId', orderId, True)
    filePath = sanitize.validString('filePath', filePath, True)

    if filePath[len(filePath)-4 : len(filePath) ] != '.zip':
        raise ValueError('filePath must end with .zip')

    apiManager.download('/path/vector/layer/order/' + orderId + '/download', filePath, token)



import geopandas as gpd
from shapely.geometry import Polygon
from geopy.distance import geodesic
import numpy as np
import pandas as pd
import math
import sys
from datetime import datetime
from shapely import geometry
import rasterio
import matplotlib.pyplot as plt
from PIL import Image
from skimage.transform import resize
import os
import subprocess
from ellipsis import sanitize
from rasterio.warp import reproject as warp, Resampling, calculate_default_transform


def recurse(f, body, listAll, extraKey = None):
    
    r = f(body)
    if listAll:
        nextPageStart = r['nextPageStart']
        while nextPageStart != None:
            body['pageStart'] = nextPageStart
            r_new = f(body)
            nextPageStart = r_new['nextPageStart']
            if 'size' in r.keys():
                r['size'] = r['size'] + r_new['size']
            if extraKey == None:
                r['result'] =  r['result'] + r_new['result']
            else:
                r['result'][extraKey] = r['result'][extraKey] +  r_new['result'][extraKey]
                
        r['nextPageStart'] = None
    return r


def stringToDate(date):
    date = sanitize.validString('date', date, True)

    try:
        d = datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%fZ")
    except:
        try:
            d = datetime.strptime(date, "%Y-%m-%d %H:%M:%S.%f")
        except:
            d = datetime.strptime(date, "%Y-%m-%d %H:%M:%S")

    return d

def dateToString(date):
    date = sanitize.validDate('date', date , True)
    
    d = date.strftime(date, "%Y-%m-%d %H:%M:%S.%fZ")
    return d

def plotRaster(raster):
    raster = sanitize.validNumpyArray('raster', raster, True)

    if len(raster.shape) != 3:
        raise ValueError('raster must have 3 dimensions')

    if raster.shape[0] != 1 and raster.shape[0] != 3:
        raise ValueError('raster must have either 1 band or 3 bands')

    

    if raster.shape[0] ==3:
        raster = np.transpose(raster, [1,2,0])
        minimum = np.min(raster)
        maximum = np.max(raster)
        if minimum == maximum:
            maximum = minimum + 1
        raster = raster - minimum
        raster = raster / (maximum-minimum)
        raster = raster * 255
        image = Image.fromarray(raster.astype('uint8'))
        image.show()

    else:
        plt.imshow(raster[0,:,:], interpolation='none')
        plt.show()
    
def plotFeatures(features):
    features = sanitize.validGeopandas('features', features, True)
    features.plot()


def chunks(l, n = 3000):
    l = sanitize.validList('l', l, True)
    n = sanitize.validInt('n', n , False)
    result = list()
    for i in range(0, len(l), n):
        result.append(l[i:i+n])
    return(result)
    

 
def cover(bounds, w):
    
    w = sanitize.validInt('w',w, True)
    bounds = sanitize.validBounds('bounds',bounds, True)

    x1 = bounds['xMin']
    y1 = bounds['yMin']
    x2 = bounds['xMax']
    y2  = bounds['yMax']

    step_y =  w/geodesic((y1,x1), (y1 - 1,x1)).meters
    parts_y = math.floor((y2 - y1)/ step_y + 1)

    y1_vec = y1 + np.arange(0, parts_y )*step_y
    y2_vec = y1 + np.arange(1, parts_y +1 )*step_y
        
    steps_x = [   w/geodesic((y,x1), (y,x1+1)).meters for y in y1_vec  ]

    parts_x = [math.floor( (x2-x1) /step +1 ) for step in steps_x ]      
        

    coords = pd.DataFrame()
    for n in np.arange(len(parts_x)):
        x1_sq = [ x1 + j*steps_x[n] for j in np.arange(0,parts_x[n]) ]
        x2_sq = [ x1 + j*steps_x[n] for j in np.arange(1, parts_x[n]+1) ]
        coords_temp = {'x1': x1_sq, 'x2': x2_sq, 'y1': y1_vec[n], 'y2':y2_vec[n]}
        coords = coords.append(pd.DataFrame(coords_temp))

    cover = [Polygon([ (coords['x1'].iloc[j] , coords['y1'].iloc[j]) , (coords['x2'].iloc[j] , coords['y1'].iloc[j]), (coords['x2'].iloc[j] , coords['y2'].iloc[j]), (coords['x1'].iloc[j] , coords['y2'].iloc[j]) ]) for j in np.arange(coords.shape[0])]
     


    coords = gpd.GeoDataFrame({'geometry': cover, 'x1':coords['x1'], 'x2':coords['x2'], 'y1':coords['y1'], 'y2':coords['y2'] })

    coords.crs = 'epsg:4326'

    return(coords)
    

    
def loadingBar(count,total):
    
    count = sanitize.validInt('count', count, True)
    total = sanitize.validInt('total', total, True)
    
    if total == 0:
        return
    else:
        percent = float(count)/float(total)*100
        sys.stdout.write("\r" + str(int(count)).rjust(3,'0')+"/"+str(int(total)).rjust(3,'0') + ' [' + '='*int(percent) + ' '*(100-int(percent)) + ']')


def reprojectRaster(r, sourceExtent, targetExtent, targetWidth, targetHeight, targetCrs, tempFolder, interpolation = 'nearest'):
    
    sanitize.validNumpyArray('r', r, True)
    targetExtent, targetCrs = sanitize.validBoundsCrsCombination(['targetExtent', 'targetCrs'], [targetExtent, targetCrs], True)
    tempFolder = sanitize.validString('tempFolder', tempFolder, True)
    targetWidth = sanitize.validInt('targetWidth', targetWidth, True)
    targetHeight = sanitize.validInt('targetHeight', targetHeight, True)
    interpolation = sanitize.validString('interpolation', interpolation, True)
    
    if not os.path.exists(tempFolder):
        raise ValueError(tempFolder + ' not found')
    if not os.path.isdir(tempFolder):
        raise ValueError(tempFolder + ' must be a path to a folder')
    if len(r.shape) !=3:
        raise ValueError('r must be 3 dimensional')
    if interpolation != 'nearest' and interpolation != 'linear':
        raise ValueError('interpolation must be either nearest or linear')
        
    dtype = r.dtype
    def mercatorBlowUp(y):
       factor = math.cos(y * 2 * math.pi/360)
       return(factor)    
    ###STEP 1
    #calculate zoom
    target_tile = geometry.Polygon([(targetExtent['xMin'],targetExtent['yMin']), (targetExtent['xMin'],targetExtent['yMax']), (targetExtent['xMax'],targetExtent['yMax']), (targetExtent['xMax'], targetExtent['yMin'])])
    target_tile = gpd.GeoDataFrame({"geometry": [target_tile]})
    target_tile.crs = targetCrs
    
    wgs_tile = target_tile.to_crs('epsg:4326')
    minx_wgs = wgs_tile.bounds['minx'].values[0]
    miny_wgs = wgs_tile.bounds['miny'].values[0]
    maxx_wgs = wgs_tile.bounds['maxx'].values[0]
    maxy_wgs = wgs_tile.bounds['maxy'].values[0]
    if minx_wgs < -180 or maxx_wgs > 180 or miny_wgs < -85 or maxy_wgs > 85:
        raise ValueError('Bounds must be within 85 degrees south and 85 degrees north')
    native_transform = rasterio.transform.from_bounds(targetExtent['xMin'], targetExtent['yMin'], targetExtent['xMax'], targetExtent['yMax'], targetWidth, targetHeight)
    source_transform = rasterio.transform.from_bounds(sourceExtent['xMin'], sourceExtent['yMin'], sourceExtent['xMax'], sourceExtent['yMax'], r.shape[2], r.shape[1])


    M = [[native_transform[0], native_transform[1]] , [native_transform[3], native_transform[4]] ]
    bias = [native_transform[2], native_transform[5]]
    Minv = np.linalg.inv(M)

    x_res, y_res = getResolution(minx_wgs, maxx_wgs, miny_wgs, maxy_wgs, targetCrs, Minv, bias)
    resolution = min(x_res, y_res)
    zoom = math.log(40000000 / (resolution*256 )) / math.log(2)
    zoom = math.floor(zoom + 1) #+ 0.85)
    
    zoom = zoom -5
    zoom = max(0, zoom)

    #####STEP 2    
    #step 1 on create raster of zoom - 4 blocks
    source_tile = geometry.Polygon([(sourceExtent['xMin'],sourceExtent['yMin']), (sourceExtent['xMin'],sourceExtent['yMax']), (sourceExtent['xMax'],sourceExtent['yMax']), (sourceExtent['xMax'], sourceExtent['yMin'])])
    source_tile = gpd.GeoDataFrame({"geometry": [source_tile]})
    source_tile.crs = 'epsg:3857'
    
    minx_merc_target = source_tile.bounds['minx'].values[0]
    miny_merc_target = source_tile.bounds['miny'].values[0]
    maxx_merc_target = source_tile.bounds['maxx'].values[0]
    maxy_merc_target = source_tile.bounds['maxy'].values[0]

    LEN = 2.003751e+07

    x_start = 2**zoom * (minx_merc_target + LEN) / (2* LEN)
    x_end = 2**zoom * (maxx_merc_target + LEN) / (2* LEN)
    y_start = 2**zoom * (miny_merc_target + LEN) / (2* LEN)
    y_end = 2**zoom * (maxy_merc_target + LEN) / (2* LEN)

    x1_osm = math.floor(x_start)
    x2_osm = math.floor(x_end)
    y1_osm = math.floor(y_start)
    y2_osm = math.floor(y_end)
    
    raster_merc_x1 = 2*LEN * x1_osm /2**zoom - LEN 
    raster_merc_x2 = 2*LEN * (x2_osm+1) /2**zoom - LEN 
    raster_merc_y1 = 2*LEN * y1_osm /2**zoom - LEN 
    raster_merc_y2 = 2*LEN * (y2_osm +1) /2**zoom - LEN 
    bounds_mercator = {'xMin': raster_merc_x1,'xMax': raster_merc_x2, 'yMin': raster_merc_y1, 'yMax': raster_merc_y2 }
    
        

    w = (x2_osm - x1_osm + 1) * 256*2**5
    h = (y2_osm - y1_osm + 1) * 256*2**5
    
    raster = np.zeros((r.shape[0], h, w), dtype = dtype)

    ###STEP 2    
    #place r on this raster
    x_frac = (x_start - x1_osm) / (x2_osm - x1_osm + 1)
    y_frac = (y2_osm + 1 - y_end) / (y2_osm - y1_osm + 1)
    
    #######33
    full_raster_tile = geometry.Polygon([(raster_merc_x1,raster_merc_y1), (raster_merc_x1, raster_merc_y2), (raster_merc_x2, raster_merc_y2), (raster_merc_x2, raster_merc_y1)])
    full_raster_tile = gpd.GeoDataFrame({"geometry": [full_raster_tile]})
    full_raster_tile.crs = 'EPSG:3857'
    
    xStart = math.floor(w * x_frac)
    xEnd = xStart + r.shape[2]
    yStart = math.floor(h * y_frac)
    yEnd = yStart + r.shape[1]
    
    
    ############################

    newW, newH = getVirtualSize(bounds_mercator, targetCrs, native_transform)
    
    
    
    if 'float' in str(dtype).lower():
        raster[:,:,:] = np.nan
    raster[:, yStart:yEnd,xStart:xEnd] = r


    ##interpolation work
    if interpolation == 'linear':
        order = 1
        resampleMethod = Resampling.bilinear
        interpolationString = '-r bilinear '
    else:
        order = 0
        resampleMethod = Resampling.nearest
        interpolationString = '-r near '
    

    ####STEP 3
    #resize

    raster = np.transpose(raster, [1,2,0])
    raster = resize(raster, ( newH, newW), anti_aliasing= False, order = order, mode = 'edge',preserve_range=True)
    raster = np.transpose(raster, [2,0,1])
    raster = raster.astype(dtype)
    saveRaster( targetFile =  os.path.join(tempFolder,'resized_temp.tif'), r = raster, extent = bounds_mercator, crs = 'EPSG:3857')



    ####step 6 gdalwarp on smaller raster
    
    extent = '-te ' + str(targetExtent['xMin']) + ' ' + str(targetExtent['yMin']) + ' ' + str(targetExtent['xMax']) + ' ' + str(targetExtent['yMax']) + " "
    outsize = '-ts ' + str(targetWidth) + ' ' + str(targetHeight) + " "
    epsgString = " -t_srs " + targetCrs + " "
    

    cmd = "gdalwarp -overwrite " + outsize + interpolationString  +  extent + " -s_srs EPSG:3857 " + epsgString + os.path.join(tempFolder,'resized_temp.tif') + " " + os.path.join(tempFolder, 'output_temp.tif')
    try:
        status = subprocess.check_output(cmd, shell=True)
    except:
        raise ValueError('No gdal instalation found')
    
    del raster
    with rasterio.open(os.path.join(tempFolder, 'output_temp.tif')) as con:
        raster = con.read()

    os.remove( os.path.join(tempFolder,'resized_temp.tif'))
    os.remove( os.path.join(tempFolder,'output_temp.tif'))

    raster = np.transpose(raster, [1,2,0])
    raster = resize(raster, ( targetHeight, targetWidth), anti_aliasing= False, order = order, mode = 'edge',preserve_range=True)
    raster = np.transpose(raster, [2,0,1])
    raster = raster.astype(dtype)



    return {'raster':raster, 'transform':native_transform, 'extent':targetExtent, 'crs':targetCrs}

def getResolution(minx_wgs, maxx_wgs, miny_wgs, maxy_wgs, crs, Minv, bias):
    STEPS = 10
    
    x_step = (maxx_wgs - minx_wgs) / STEPS
    y_step = (maxy_wgs - miny_wgs) / STEPS

    
    pixelPositions = {}
    n=0
    m=0
    for n in [0,1]:
        for m in [0,1]:
            if m != 1 or n !=1:
                points = [ geometry.Point((minx_wgs + (i+n) * x_step, miny_wgs + (j+m) * y_step))  for i in np.arange(STEPS) for j in np.arange(STEPS)]
                df_wgs = gpd.GeoDataFrame({'geometry':points})
                df_wgs.crs = 'epsg:4326'
                df_native = df_wgs.to_crs(crs)
                pixelPositions[str(n) + '_' + str(m)] = np.array([ np.matmul(Minv , [point.x - bias[0], point.y -bias[1]] ) for point in df_native['geometry'].values])

    distances_x = [geodesic( (miny_wgs + j * y_step, minx_wgs + i * x_step), (miny_wgs + j * y_step, minx_wgs + (i+1) * x_step)).m  for i in np.arange(STEPS) for j in np.arange(STEPS)]
    distances_y = [geodesic( (miny_wgs + j * y_step, minx_wgs + i * x_step), (miny_wgs + (j+1) * y_step, minx_wgs + i * x_step)).m  for i in np.arange(STEPS) for j in np.arange(STEPS)]
    dif_pixel_x  = pixelPositions['1_0'] - pixelPositions['0_0']
    dif_pixel_y  = pixelPositions['0_1'] - pixelPositions['0_0']
    
    distances_pixel_x = [math.sqrt( p[0]**2 + p[1]**2) for p in dif_pixel_x]
    distances_pixel_y = [math.sqrt( p[0]**2 + p[1]**2) for p in dif_pixel_y]

    x_res = np.divide(distances_x, distances_pixel_x)
    y_res = np.divide(distances_y, distances_pixel_y)
    
    x_res = np.min(x_res)
    y_res = np.min(y_res)
    
    return x_res, y_res

def getVirtualSize(bounds_mercator, targetCrs, native_transform):
    M = [[native_transform[1], native_transform[2]] , [native_transform[4], native_transform[5]] ]
    bias = [native_transform[0], native_transform[3]]
    
    Minv = np.linalg.inv(M)
    
    #convert bounds_tile to raster crs
    lowerLeft = transformPoint( (bounds_mercator['xMin'],bounds_mercator['yMin']), 'EPSG:3857', targetCrs)
    upperLeft = transformPoint( (bounds_mercator['xMin'],bounds_mercator['yMax']), 'EPSG:3857', targetCrs)
    lowerRight = transformPoint( (bounds_mercator['xMax'],bounds_mercator['yMin']), 'EPSG:3857', targetCrs)
    
    #calculate pixels using inverse transform of 3 points
    lowerLeft_pixel = np.matmul(Minv , [lowerLeft[0] - bias[0], lowerLeft[1] -bias[1]] ) 
    upperLeft_pixel = np.matmul(Minv , [upperLeft[0] - bias[0], upperLeft[1] -bias[1]] ) 
    lowerRight_pixel = np.matmul(Minv , [lowerRight[0] - bias[0], lowerRight[1] -bias[1]] ) 
    
    #calculate pixel distance
    w = math.sqrt( (lowerLeft_pixel[0] - lowerRight_pixel[0] )**2 + (lowerLeft_pixel[1] - lowerRight_pixel[1] )**2 )    
    h = math.sqrt( (lowerLeft_pixel[0] - upperLeft_pixel[0] )**2 + (lowerLeft_pixel[1] - upperLeft_pixel[1] )**2 )    
    
    w = math.floor(w+1)
    h = math.floor(h+1)
    
    return w,h
    

def transformPoint( point, sourceCrs, targetCrs):
    try:
        point = geometry.Point(point)
    except:
        raise ValueError('point must be a tuple with two floats')
    df = gpd.GeoDataFrame({'geometry':[point]})    
    try:
        df.crs = sourceCrs
    except:
        raise ValueError('sourceCrs not a valid crs')
    try:
        df = df.to_crs(targetCrs)
    except:
        raise ValueError('targetCrs not a valid crs')
    x = df.bounds['minx'].values[0]
    y = df.bounds['miny'].values[0]

    return(x,y)

#funciton to store image given bounds, and raster and crs
def saveRaster(targetFile, r, crs, extent = None, transform = None):
    if type(extent) != type(None):
        sanitize.validBoundsCrsCombination([extent, crs], [extent, crs], True)
        
    r = sanitize.validNumpyArray('r', r, True)
    
    if len(r.shape) != 3:
        raise ValueError('r must be 3 dimensional')

    w = r.shape[2]
    h = r.shape[1]
    if transform == None:
        transform = rasterio.transform.from_bounds(extent['xMin'], extent['yMin'], extent['xMax'], extent['yMax'], w, h)
    con = rasterio.open(targetFile , 'w', driver='Gtiff',compress="lzw",
                    height = h, width = w,
                    count= r.shape[0], dtype=r.dtype,
                    crs=  crs ,
                    transform=transform)
    
    con.write(r)
    con.close()



from ellipsis.path.raster.timestamp.root import add
from ellipsis.path.raster.timestamp.root import edit
from ellipsis.path.raster.timestamp.root import delete
from ellipsis.path.raster.timestamp.root import activate
from ellipsis.path.raster.timestamp.root import getBounds
from ellipsis.path.raster.timestamp.root import getAggregatedData
from ellipsis.path.raster.timestamp.root import getRaster
from ellipsis.path.raster.timestamp.root import getDownsampledRaster
from ellipsis.path.raster.timestamp.root import trash
from ellipsis.path.raster.timestamp.root import recover

from ellipsis.path.raster.timestamp import upload
from ellipsis.path.raster.timestamp import order





from uuid import UUID
import numpy as np
import datetime
import json
import geopandas as gpd
import pandas as pd
from shapely import geometry

def validUuid(name, value, required):
    if not required and type(value) == type(None):
        return
    try:
        UUID(value, version=2)
    except:
        raise ValueError(name + ' must be of type string and be a uuid')
    return(value)


def validUuidArray(name, value, required):
    if not required and type(value) == type(None):
        return

    try:
        value = list(value)        
    except:
        raise ValueError(name + ' must be an iterable')

    for uuid in value:
        try:
            UUID(uuid, version=2)
        except:
            raise ValueError(name + ' must be a list of uuids')    
    return(value)



def validString(name, value, required):
    if not required and type(value) == type(None):
        return

    if type(value) != type('x'):
        raise ValueError(name + 'must be of type string')
    return(value)

def validShapely(name, value, required):
    if not required and type(value) == type(None):
        return

    if not 'shapely' in str(type(value)):
        raise ValueError(name + 'must be a shapely geometry')
    return(value)


def validBoundsCrsCombination(name, value, required):
    if not required and type(value) == type(None):
        return

    bounds = value[0]
    crs = value[1]
    
    value = bounds
    keys = ['xMin', 'xMax', 'yMin', 'yMax']
    if type(value) != type({}):
        raise ValueError(name[0] + ' must be a dictionary with keys ' ' '.join(keys) +
                         ' whose values must be of type float')

    for key in keys:
        if not key in value.keys() or (not 'float' in str(type(value[key])) and not 'int' in str(type(value[key]))):
            raise ValueError(name[0] + ' must be a dictionary with keys ' ' '.join(
                keys) + ' whose values must be of type float')
        value[key] = float(value[key])


    if type(crs) != type('x'):
        raise ValueError('crs must be of type string')

    target_tile = geometry.Polygon([(bounds['xMin'],bounds['yMin']), (bounds['xMin'],bounds['yMax']), (bounds['xMax'],bounds['yMax']), (bounds['xMax'], bounds['yMin'])])
    sh = gpd.GeoDataFrame({'geometry':[target_tile]})    
    try:
        sh.crs = crs
    except:
        raise ValueError(name[0] + ' does not fit with the projection given by ' + name[1])

    return bounds, crs        

def validImage(name, value, required):
    if not required and type(value) == type(None):
        return

    if type(value) != type(np.zeros((2,2))):
        raise ValueError(name + ' must be a numpy array with either 2 or 3 dimensions')
    if len(value.shape) != 2  and len(value.shape) !=3 :
        raise ValueError(name + ' must be a numpy array with either 2 or 3 dimensions')
    if len(value.shape) ==3 and value.shape[2] !=3 and value.shape[2] !=1:
        raise ValueError(name + ' must have either 1 or 3 bands')

    return value

def validDataframe(name, value, required):
    if not required and type(value) == type(None):
        return

    if type(value) != type(pd.DataFrame()):
        raise ValueError(name + ' must be a pandas dataframe')
    value = value.copy()
    return value
        

def validNumpyArray(name, value, required):
    if not required and type(value) == type(None):
        return

    if type(value) != type(np.zeros((2,2) )):
        raise ValueError(name + 'must be a geopandas dataframe')

    return value


def validList(name, value, required):
    if not required and type(value) == type(None):
        return

    try:
        value = list(value)

    except:
        raise ValueError(name + ' must be an iterable')
    return value


def validGeopandas(name, value, required):
    if not required and type(value) == type(None):
        return

    if type(value) != type(gpd.GeoDataFrame()):
        raise ValueError(name + 'must be a geopandas dataframe')

    value = value.copy()        
    if str(type(value.crs)) == str(type(None)) and min(value.bounds['minx']) > -180  and max(value.bounds['maxx']) < 180 and min(value.bounds['miny']) > -90  and max(value.bounds['maxy']) < 90:
        print('assuming WGS84 coordinates for ' + name)
        value.crs = {'init': 'epsg:4326'}
    elif str(type(value.crs)) == str(type(None)):
        raise ValueError('Please provide CRS for ' + name)
    else:
        value = value.to_crs({'init': 'epsg:4326'})

    if 'id' in value.columns:
        del value['id']
    if 'userId' in value.columns:
        del value['userId']
    if 'attribution' in value.columns:
        del value['attribution']
    if 'radius' in value.columns:
        del value['radius']
    if 'color' in value.columns:
        del value['color']
        
    return(value)


def validBool(name, value, required):
    if not required and type(value) == type(None):
        return

    if not 'bool' in str(type(value)):
        raise ValueError(name + ' must be of type boolean')
    value = bool(value)
    return(value)


def validObject(name, value, required):
    if not required and type(value) == type(None):
        return


    try:
        json.dumps(value)
    except:
        raise ValueError(name + ' must be json serializable')
    return value

def validInt(name, value, required):
    if not required and type(value) == type(None):
        return

    if not 'int' in str(type(value)).lower():
        raise ValueError(name + ' must be of type int')

    value = int(value)
    return(value)


def validFloat(name, value, required):
    if not required and type(value) == type(None):
        return

    if not 'int' in str(type(value)).lower() and not 'float' in str(type(value)).lower():
        raise ValueError(name + ' must be of type float')

    value = float(value)
    return(value)


def validBounds(name, value, required):
    if not required and type(value) == type(None):
        return
    keys = ['xMin', 'xMax', 'yMin', 'yMax']
    if type(value) != type({}):
        raise ValueError(name + ' must be a dictionary with keys ' ' '.join(keys) +
                         ' whose values must be of type float')

    for key in keys:
        if not key in value.keys() or (not 'float' in str(type(value[key])) and not 'int' in str(type(value[key]))):
            raise ValueError(name + ' must be a dictionary with keys ' ' '.join(
                keys) + ' whose values must be of type float')
        value[key] = float(value[key])

    if value['xMin'] > 180 or value['xMin'] < -180:
        raise ValueError('xMin must be between -180 and 180')
    if value['xMax'] > 180 or value['xMax'] < -180:
        raise ValueError('xMax must be between -180 and 180')
    if value['yMin'] > 85 or value['yMin'] < -85:
        raise ValueError('yMin must be between -85 and 85')
    if value['yMax'] > 85 or value['yMax'] < -85:
        raise ValueError('yMax must be between -180 and 180')
    if value['yMax'] <= value['yMin']:
        raise ValueError('yMax must be strictly greater than yMin')
    if value['xMax'] <= value['xMin']:
        raise ValueError('xMax must be strictly greater than xMin')


    return(value)


def validStringArray(name, value, required):
    if not required and type(value) == type(None):
        return

    try:
        value = list(value)        

    except:
        raise ValueError(name + ' must be an iterable')

    types = np.array([type(x) != type('x') for x in value])

    if np.sum(types) > 0:
        raise ValueError(name + ' must be a list of strings')

    return value


def validFloatArray(name, value, required):
    if not required and type(value) == type(None):
        return

    try:
        value = list(value)

    except:
        raise ValueError(name + ' must be an iterable')

    types = np.array([(not 'float' in str(type(x))) and (
        not 'int' in str(type(x))) for x in value])

    if np.sum(types) > 0:
        raise ValueError(name + ' must be a list of strings')

    value = [float(x) for x in value]

    return value


def validIntArray(name, value, required):
    if not required and type(value) == type(None):
        return

    try:
        value = list(value)

    except:
        raise ValueError(name + ' must be an iterable')

    types = np.array([not 'int' in str(type(x)) for x in value])

    if np.sum(types) > 0:
        raise ValueError(name + ' must be a list of strings')

    value = [int(x) for x in value]

    return value


def validDate(name, value, required):

    if not required and type(value) == type(None):
        return

    if type(value) != type(datetime.datetime.now()):
        raise ValueError(name + ' must be of type datetime')

    value = value.strftime('%Y-%m-%dT%H:%M:%S')
    return value






from ellipsis.path.vector.layer.root import add, edit, delete, recover, archive, getBounds, getChanges, getFeaturesByExtent, getFeaturesByIds, listFeatures, editFilter
from ellipsis.path.vector.layer import featureProperty
from ellipsis.path.vector.layer import style
from ellipsis.path.vector.layer import feature
from ellipsis.path.vector.layer import upload
from ellipsis.path.vector.layer import order





from ellipsis.path.root import searchFolder, searchRaster, favorite, unfavorite, editPublicAccess, searchVector, listMaps, listFolders, delete, editMetadata, move, rename, trash, recover, delete, get, add
from ellipsis.path import hashtag
from ellipsis.path import invite
from ellipsis.path import member
from ellipsis.path import raster
from ellipsis.path import vector

